import boto3
from chalice import Cron
import copy
import cron_descriptor
import logging
import os
from typing import Callable, Optional
from ...decorators import Decorators
from .envs import Envs
from .misc_utils import memoize

logging.basicConfig()
logger = logging.getLogger(__name__)


class Checks:

    def __init__(self, check_setup, envs: Envs):
        self._check_setup_raw = check_setup
        self._check_setup = copy.deepcopy(check_setup)
        self._envs = envs

    def get_checks_raw(self) -> dict:
        """
        Returns a dictionary containing the pristine, original check_setup.json file contents.
        """
        return self._check_setup_raw

    @memoize
    def _get_checks(self) -> dict:
        checks = self._check_setup
        for check_key in checks.keys():
            checks[check_key]["name"] = check_key
            checks[check_key]["group"] = checks[check_key]["group"]
        lambdas = self.get_annotated_lambdas()
        self._annotate_checks_for_dependencies(checks)
        self._annotate_checks_with_schedules_from_lambdas(checks, lambdas)
        self._annotate_checks_with_kwargs_from_decorators(checks)
        return checks

    @memoize
    def get_checks(self, env: str) -> dict:
        """
        Returns a dictionary containing all checks, annotated with various info,
        e.g. the (cron) schedule from the associated lambdas, check kwargs, and
        associated actions, from the check function decorators; filtered by the
        given env name. Cached on the first call, except for the filtering part.
        """
        return self._filter_checks_by_env(self._get_checks(), env)

    @memoize
    def get_checks_grouped(self, env: str) -> list:
        """
        Like get_checks but returns the checks grouped by their group names.
        """
        grouped_checks = []
        checks = self.get_checks(env)
        for check_name in checks:
            check_item = checks[check_name]
            check_item_group = check_item["group"]
            found = False
            for grouped_check in grouped_checks:
                if grouped_check["group"] == check_item_group:
                    grouped_check["checks"].append(check_item)
                    found = True
                    break
            if not found:
                grouped_checks.append({"group": check_item_group, "checks": [check_item]})
        return grouped_checks

    @memoize
    def get_checks_grouped_by_schedule(self, env: str) -> list:
        """
        Like get_checks_grouped but groups by schedule (i.e. by scheduling lambdas).
        """
        def create_schedule_group_title(name: str):
            """
            Pretty lame title-ifying the groiup name for display purposes, based
            on what we know about the currently existing scheduling lambda names.
            """
            name = name.replace("Min", " Minute")
            if name.endswith("Checks"):
                return name[:len(name) - len("Checks") - 0]
            elif name.endswith("Checks1"):
                return name[:len(name) - len("Checks1") - 0] + " I"
            elif name.endswith("Checks2"):
                return name[:len(name) - len("Checks2") - 0] + " II"
            elif name.endswith("Checks3"):
                return name[:len(name) - len("Checks3") - 0] + " III"
            elif name.endswith("Checks4"):
                return name[:len(name) - len("Checks4") - 0] + " IV"
            elif name.endswith("Checks5"):
                return name[:len(name) - len("Checks5") - 0] + " V"
            else:
                return name
        grouped_checks = []
        lambdas = self.get_annotated_lambdas()
        for lambda_item in lambdas:
            schedule_group_name = create_schedule_group_title(lambda_item["lambda_name"])
            if isinstance(lambda_item.get("lambda_checks"), list) and len(lambda_item["lambda_checks"]) > 0:
                for lambda_item_check in lambda_item["lambda_checks"]:
                    check = self.get_check(env, lambda_item_check.get("check_name"))
                    if check:
                        group = [group for group in grouped_checks if group["group"] == schedule_group_name]
                        if len(group) == 0:
                            group = {
                                "group": schedule_group_name,
                                "checks": []
                            }
                            grouped_checks.append(group)
                        else:
                            group = group[0]
                        group["checks"].append(check)
        return grouped_checks

    @memoize
    def get_check(self, env: str, check: str) -> Optional[dict]:
        """
        Returns the check for the given check name; filtered by the given env name.
        If it turns out the check name is really an action then return its info;
        there is a 'type' property set to 'check' or 'action indicating which it is.
        """
        checks = self.get_checks(env)
        for check_key in checks.keys():
            if check_key == check:
                return {"type": "check", **checks[check_key]}
        return self._get_action(env, check)

    def _get_action_checks(self, env: str, action: str) -> list:
        """
        Returns the list of checks associated with the given action.
        """
        action_checks = []
        checks = self.get_checks(env)
        for check in checks.values():
            if check.get("registered_action"):
                if check["registered_action"].get("name") == action:
                    action_checks.append(check)
        return action_checks

    def _get_action(self, env: str, action: str) -> Optional[dict]:
        checks = self.get_checks(env)
        for check_key in checks.keys():
            check = checks[check_key]
            if check.get("registered_action") and check["registered_action"].get("name") == action:
                action_checks = self._get_action_checks(env, action)
                action_title = " ".join(action.split("_")).title()
                return {"type": "action", **check["registered_action"], "checks": action_checks, "title": action_title}
        return None

    def _filter_checks_by_env(self, checks: dict, env: str) -> dict:
        """
        Returns the given checks filtered by the given env name.
        """
        if not env:
            return checks
        checks_for_env = {}
        for check_key in checks.keys():
            included = False
            if checks[check_key].get("schedule"):
                for check_schedule_key in checks[check_key]["schedule"].keys():
                    for check_env_key in checks[check_key]["schedule"][check_schedule_key].keys():
                        if check_env_key == "all" or self._envs.is_same_env(check_env_key, env):
                            included = True
                            checks_for_env[check_key] = checks[check_key]
            else:
                # If no schedule section (which has the env section) then include it.
                included = True
                checks_for_env[check_key] = checks[check_key]
            if not included:
                if isinstance(checks[check_key].get("display"), list):
                    for check_env_key in checks[check_key]["display"]:
                        if check_env_key == "all" or self._envs.is_same_env(check_env_key, env):
                            checks_for_env[check_key] = checks[check_key]
        return checks_for_env

    def _annotate_checks_for_dependencies(self, checks: dict) -> None:
        """
        Annotates the give checks dictionary with "referrer" information related to the
        check "dependencies". A check may have a list of dependencies, and for general
        info (in the UI) and troubleshooting purposes, it would be useful to also 
        have the inverse relationship, i.e. for each check which is part of another
        check dependency, we create a "referrers" list with the names of any checks
        which depend on (refer to) the check, i.e. which have it in its dependencies
        list. This is intended to be called once and cached (see: get_checks).
        """
        def get_check(check_name: str, env: str) -> Optional[dict]:
            """
            Returns the check from the (outer) checks dictionary with the given name;
            we can't use the top-level get_checks because that one calls into this,
            which would result in infinite recursion; this is a startup/bootstrap activity.
            """
            for check_key in self._filter_checks_by_env(checks, env):
                if check_key == check_name:
                    return checks[check_key]
            return None

        for check_name in checks:
            check = checks[check_name]
            if isinstance(check.get("schedule"), dict):
                check_schedules = check["schedule"]
                for check_schedule_name in check_schedules:
                    check_schedule = check_schedules[check_schedule_name]
                    if isinstance(check_schedule, dict):
                        for check_schedule_key in check_schedule:
                            check_schedule_item = check_schedule[check_schedule_key]
                            if isinstance(check_schedule_item, dict):
                                check_schedule_env_name = check_schedule_key
                                if isinstance(check_schedule_item.get("dependencies"), list):
                                    for check_dependency_name in check_schedule_item["dependencies"]:
                                        check_dependency = get_check(check_dependency_name, check_schedule_env_name)
                                        if check_dependency:
                                            if not isinstance(check_dependency.get("referrers"), list):
                                                check_dependency["referrers"] = [check_name]
                                            elif check_name not in check_dependency["referrers"]:
                                                check_dependency["referrers"].append(check_name)

    @staticmethod
    def _get_stack_name() -> str:
        """
        Returns our AWS defined stack name, as specified by the STACK_NAME environment variable.
        """
        return os.environ.get("STACK_NAME")

    def _get_stack_template(self, stack_name: str = None) -> dict:
        """
        Returns our AWS stack template, for our defined AWS stack.
        """
        if not stack_name:
            stack_name = self._get_stack_name()
            if not stack_name:
                return {}
        boto_cloudformation = boto3.client('cloudformation')
        return boto_cloudformation.get_template(StackName=stack_name)

    @staticmethod
    def _get_lambdas_from_template(stack_template: dict) -> list:
        """
        Returns the list of lambda names and associated info for our defined AWS stack.
        """
        lambda_definitions = []
        stack_template = stack_template["TemplateBody"]["Resources"]
        for resource_key in stack_template:
            resource_type = stack_template[resource_key]["Type"]
            if resource_type == "AWS::Lambda::Function":
                lambda_name = resource_key
                lambda_properties = stack_template[lambda_name]["Properties"]
                lambda_code_s3_bucket = lambda_properties["Code"]["S3Bucket"]
                lambda_code_s3_bucket_key = lambda_properties["Code"]["S3Key"]
                lambda_handler = lambda_properties["Handler"]
                lambda_definitions.append({
                    "lambda_name": lambda_name,
                    "lambda_code_s3_bucket": lambda_code_s3_bucket,
                    "lambda_code_s3_bucket_key": lambda_code_s3_bucket_key,
                    "lambda_handler": lambda_handler
                })
        return lambda_definitions

    @staticmethod
    def _annotate_lambdas_with_schedules_from_template(lambdas: list, stack_template: dict) -> list:
        """
        Annotates and returns the given AWS lambdas list with
        the (cron) schedules from the given associated AWS stack template.
        """
        def is_cron_schedule_never(cron: str) -> bool:
            try:
                cron_split = cron.split(' ')
                cron = Cron(*cron_split)
                # Just check the most obvious things.
                # We are known to only use Feb 31 to indicate Never.
                if cron.month.isnumeric():
                    month = int(cron.month)
                    if month < 1 or month > 12:
                        return True
                else:
                    month = None
                if cron.day_of_month.isnumeric():
                    day_of_month = int(cron.day_of_month)
                    if day_of_month < 1 or day_of_month > 31:
                        return True
                    if month == 2 and day_of_month > 29:
                        return True
                    if month in [4, 6, 9, 11] and day_of_month > 30:
                        return True
            except Exception:
                pass
            return False

        stack_template = stack_template["TemplateBody"]["Resources"]
        for resource_key in stack_template:
            resource_type = stack_template[resource_key]["Type"]
            if resource_type == "AWS::Events::Rule":
                event_name = resource_key
                event_properties = stack_template[event_name]["Properties"]
                event_schedule = event_properties["ScheduleExpression"]
                if event_schedule:
                    event_targets = event_properties["Targets"]
                    for event_target in event_targets:
                        event_target = dict(event_target)
                        event_target_arn = dict(event_target["Arn"])
                        event_target_function_arn = event_target_arn["Fn::GetAtt"]
                        if len(event_target_function_arn) == 2 and "Arn" in event_target_function_arn:
                            if event_target_function_arn[0] == "Arn":
                                event_target_function_name = event_target_function_arn[1]
                            else:
                                event_target_function_name = event_target_function_arn[0]
                            if event_target_function_name:
                                for la in lambdas:
                                    if la["lambda_name"] == event_target_function_name:
                                        event_schedule = str(event_schedule).replace("cron(", "").replace(")", "")
                                        la["lambda_schedule"] = str(event_schedule)
                                        if is_cron_schedule_never(event_schedule):
                                            cron_description = "Never"
                                        else:
                                            cron_description = cron_descriptor.get_description(str(event_schedule))
                                            if cron_description.startswith("At "):
                                                cron_description = cron_description[3:]
                                            cron_description = cron_description + " (UTC)"
                                        la["lambda_schedule_description"] = cron_description
        return lambdas

    @staticmethod
    def _annotate_lambdas_with_function_metadata(lambdas: list) -> list:
        """
        Annotates and returns the given AWS lambdas list with various AWS lambda metadata,
        e.g. the function name and ARN, code size, role, and description.
        """
        boto_lambda = boto3.client("lambda")
        # lambda_functions = boto_lambda.list_functions()["Functions"]
        lambda_functions = Checks._get_all_lambda_functions()
        for lambda_function in lambda_functions:
            lambda_function_handler = lambda_function["Handler"]
            for la in lambdas:
                if la["lambda_handler"] == lambda_function_handler:
                    la["lambda_function_name"] = lambda_function["FunctionName"]
                    la["lambda_function_arn"] = lambda_function["FunctionArn"]
                    la["lambda_code_size"] = lambda_function["CodeSize"]
                    la["lambda_modified"] = lambda_function["LastModified"]
                    la["lambda_description"] = lambda_function["Description"]
                    la["lambda_role"] = lambda_function["Role"]
                    #
                    # Look for the real modified time which may be in the tag if we ever did a manual
                    # reload of the lambda which will do its job by making an innocuous change to the
                    # lambda (its description) but which also has the effect of changing its modified
                    # time, so that process also squirrels away the real lambda modified time in a
                    # tag called last_modified. See the reload_lambda function for details of this.
                    try:
                        lambda_function_tags = boto_lambda.list_tags(Resource=lambda_function["FunctionArn"])["Tags"]
                        lambda_modified = lambda_function_tags.get("last_modified")
                        if lambda_modified:
                            la["lambda_modified"] = lambda_modified
                    except Exception as e:
                        logger.warning(f"Exception getting AWS lambdas info: {e}")
                        pass
        return lambdas

    @staticmethod
    def _annotate_lambdas_with_check_setup(lambdas: list, checks: dict) -> list:
        """
        Annotates and returns the given AWS lambdas list with info from the given checks.
        """
        if not checks or not isinstance(checks, dict):
            return lambdas
        for check_name in checks:
            check_item = checks[check_name]
            check_item_schedule = check_item.get("schedule")
            if check_item_schedule:
                for check_item_schedule_name in check_item_schedule.keys():
                    for lambda_item in lambdas:
                        if (lambda_item["lambda_handler"] == check_item_schedule_name
                                or lambda_item["lambda_handler"] == "app." + check_item_schedule_name):
                            if not lambda_item.get("lambda_checks"):
                                lambda_item["lambda_checks"] = []
                            lambda_item["lambda_checks"].append({
                                "check_title": check_item.get("title"),
                                "check_name": check_name,
                                "check_group": check_item.get("group")
                            })
        for lambda_item in lambdas:
            if lambda_item.get("lambda_checks"):
                lambda_item["lambda_checks"].sort(key=lambda item: f"{item['check_group']}.{item['check_name']}")
        return lambdas

    @memoize
    def _get_annotated_lambdas(self) -> dict:
        stack_name = self._get_stack_name()
        stack_template = self._get_stack_template(stack_name)
        lambdas = self._get_lambdas_from_template(stack_template)
        lambdas = self._annotate_lambdas_with_schedules_from_template(lambdas, stack_template)
        lambdas = self._annotate_lambdas_with_function_metadata(lambdas)
        lambdas = self._annotate_lambdas_with_check_setup(lambdas, self._check_setup)
        return lambdas

    @memoize
    def get_annotated_lambdas(self, env: Optional[str] = None) -> dict:
        """
        Returns the dictionary of all AWS lambdas for our defined stack.
        """
        return self._filter_lambdas_by_env(self._get_annotated_lambdas(), env)

    def _filter_lambdas_by_env(self, lambdas: list, env: Optional[str] = None) -> list:
        """
        Filters the given list of lambda info by the given environment (i.e. just the
        list of checks within each lambda info record), and returns the resultant list.
        The given list of lambda info is NOT changed (i.e. makes copies if modified).
        """
        if not env:
            return lambdas
        filtered_lambdas = []
        for lambda_item in lambdas:
            lambda_checks = lambda_item.get("lambda_checks")
            if isinstance(lambda_checks, list) and len(lambda_checks) > 0:
                lambda_item = copy.deepcopy(lambda_item)
                lambda_checks_filtered = []
                for lambda_check in lambda_checks:
                    check = self.get_check(env, lambda_check.get("check_name"))
                    if check:
                        lambda_checks_filtered.append(lambda_check)
                lambda_item["lambda_checks"] = lambda_checks_filtered
            filtered_lambdas.append(lambda_item)
        return filtered_lambdas

    @staticmethod
    def _annotate_checks_with_schedules_from_lambdas(checks: dict, lambdas: dict) -> None:
        """
        Annotates the given checks with the (cron) schedule from the given associated AWS lambdas.
        """
        for check_name in checks:
            check_item = checks[check_name]
            check_schedule = check_item.get("schedule")
            if check_schedule:
                for check_schedule_name in check_schedule.keys():
                    for la in lambdas:
                        if (la["lambda_handler"] == check_schedule_name
                                or la["lambda_handler"] == "app." + check_schedule_name):
                            check_schedule[check_schedule_name]["cron"] = la["lambda_schedule"]
                            check_schedule[check_schedule_name]["cron_description"] = la["lambda_schedule_description"]

    @memoize
    def get_registry(self) -> dict:
        """
        Returns the checks registry dictionary which was set up by the @check_function
        and @action_function decoratros (deifned in decorators.py).
        """
        registry = Decorators.get_registry()
        registered_checks = [check for check in registry if registry[check]["kind"] == "check"]
        registered_actions = [check for check in registry if registry[check]["kind"] == "action"]
        for check_name in registered_checks:
            check = registry[check_name]
            # A check may have at most one associated action;
            # but the same action may be associated with more than
            # one check; setup the latter part of that relationship here.
            associated_action_name = check.get("action")
            if associated_action_name:
                action = registry[associated_action_name]
                if not action.get("checks"):
                    action["checks"] = []
                action["checks"].append(check_name)
        return registry

    def _annotate_checks_with_kwargs_from_decorators(self, checks: dict) -> None:
        """
        Annotates the given checks with kwargs info from the check functions decorators.
        """
        # Decorators.get_registry() is a dictionary keyed by (unique) decorator function
        # name; the value of each key is an object contain the args, kwargs fields, as well
        # as other (file, module, asssociated action, etc) metadata for display purposes.
        def get_base_module_name(module: str) -> str:
            return module.split(".")[-1:][0] if module else module
        checks_registry = self.get_registry()
        registered_checks = [check for check in checks_registry if checks_registry[check]["kind"] == "check"]
        if not registered_checks:
            return
        for check_name in checks:
            check_item = checks[check_name]
            for check_function_name in registered_checks:
                if check_name == check_function_name:
                    registered_check = checks_registry[check_function_name]
                    check_item["registered_file"] = registered_check.get("file")
                    check_item["registered_line"] = registered_check.get("line")
                    check_item["registered_module"] = registered_check.get("module")
                    check_item["registered_package"] = registered_check.get("package")
                    check_item["registered_github_url"] = registered_check.get("github_url")
                    registered_check_kwargs = registered_check.get("kwargs")
                    if registered_check_kwargs:
                        check_item["registered_kwargs"] = registered_check_kwargs
                    registered_check_args = registered_check.get("args")
                    if registered_check_args:
                        check_item["registered_args"] = registered_check_args
                    # Get any associated action; this is from the (new as of early December 2022)
                    # action property specified in the @check_function decorator for the check
                    # function; this is used only for informational purposes in the UI. 
                    registered_action_function_name = registered_check.get("action")
                    if registered_action_function_name:
                        registered_action = checks_registry.get(registered_action_function_name)
                        if registered_action and registered_action.get("kind") == "action":
                            check_item["registered_action"] = {
                                "name": registered_action.get("name"),
                                "file": registered_action.get("file"),
                                "line": registered_action.get("line"),
                                "module": get_base_module_name(registered_action.get("module")),
                                "registered_module": registered_action.get("module"),
                                "package": registered_action.get("package"),
                                "github_url": registered_action.get("github_url"),
                                "checks": registered_action.get("checks")
                            }

    @staticmethod
    def _get_all_lambda_functions() -> list:
        """
        Returns a list of objects from boto3 (lambda.list_functions) for each lambda
        function within our stack as defined by _get_stack_name (i.e. from the global
        STACK_NAME environment variable, and where each lambda also has the STACK_NAME
        environment variable defined for it).
        """
        stack_name = Checks._get_stack_name()
        def lambda_filter(lambda_function: dict) -> bool:
            if lambda_function:
                lambda_function_environment = lambda_function.get("Environment")
                if lambda_function_environment:
                    lambda_function_environment_variables = lambda_function_environment.get("Variables")
                    if lambda_function_environment_variables:
                        lambda_function_stack_name = lambda_function_environment_variables.get("STACK_NAME")
                        return lambda_function_stack_name == stack_name
            return False
        boto_lambda = boto3.client("lambda")
        results = []
        marker = None
        while True:
            if marker:
                lambda_functions = boto_lambda.list_functions(Marker=marker)
            else:
                lambda_functions = boto_lambda.list_functions()
            if not lambda_functions or not lambda_functions.get("Functions"):
                break
            for lambda_function in lambda_functions["Functions"]:
                if lambda_filter(lambda_function):
                    results.append(lambda_function)
            marker = lambda_functions.get("NextMarker")
            if not marker:
                break
        return results

    def cache_clear(self) -> None:
        self._get_checks.cache_clear()
        self.get_checks.cache_clear()
        self.get_checks_grouped.cache_clear()
        self.get_checks_grouped_by_schedule.cache_clear()
        self.get_check.cache_clear()
        self._get_annotated_lambdas.cache_clear()
        self.get_annotated_lambdas.cache_clear()
        self.get_registry.cache_clear()



# AWS Cognito authentication support functions.

from cryptography.hazmat.backends.openssl.rsa import _RSAPublicKey as RSAPublicKey
import jwt as jwtlib
from jwt import PyJWKClient
import requests
import os
from typing import Tuple
from typing_extensions import TypedDict
from dcicutils.common import REGION as AWS_REGION
from foursight_core.react.api.encoding_utils import base64_encode
from foursight_core.react.api.envs import Envs
from foursight_core.react.api.gac import Gac
from foursight_core.react.api.jwt_utils import jwt_encode
from foursight_core.react.api.misc_utils import get_request_domain, get_request_origin, memoize


AWS_COGNITO_SERVICE_BASE_URL = f"https://cognito-idp.{AWS_REGION}.amazonaws.com"

KEY_REGION = "region"
KEY_DOMAIN = "domain"
KEY_USER_POOL_ID = "userpool"
KEY_CLIENT_ID = "client"
KEY_SCOPES = "scope"
KEY_CONNECTIONS = "connections"
KEY_CONFIG_URL = "config"
KEY_CALLBACK_URL = "callback"


class CognitoConfig(TypedDict):
    region: str
    domain: str
    userpool: str
    client: str
    scope: list
    connections: list
    config: str
    callback: str


def get_cognito_oauth_config(request: dict) -> CognitoConfig:
    """
    Returns all necessary configuration info for our AWS Cognito authentication server.
    Retrieved via (first) either environment variables or (second) via AWS Secrets Manager.
    :param request: Dictionary with HTTP request an backend API call (e.g. /api/reactapi/cognito_config).
    :returns: Dictionary with AWS Cognito configuration info.
    """
    config = _get_cognito_oauth_config_basic()
    #
    # We make this set-able via environment variable because it can be useful for running locally
    # in "cross-origin mode", where the React UI is running on its own on localhost:3000 (via npm)
    # and the Foursight app is running locally on localhost:8000, so that live React code updates can
    # be made and seen immediately on localhost:3000. In this case we set FOURSIGHT_COGNITO_CALLBACK
    # to localhost:3000, rather than to what the default would have been, localhost:8000.
    #
    config[KEY_CALLBACK_URL] = os.environ.get("FOURSIGHT_COGNITO_CALLBACK",
                                              f"{get_request_origin(request)}/api/react/cognito/callback")
    return config


@memoize
def _get_cognito_oauth_config_basic() -> CognitoConfig:
    """
    Returns basic configuration info for our AWS Cognito authentication server.
    Retrieved via either environment variables of AWS Secrets Manager.
    :returns: Dictionary with basic AWS Cognito configuration info.
    """
    #
    # This is separated from get_cognito_oauth_config just so we can memoize these most commonly used
    # configuration values (as the dictionary argument required for that function is not memoize-able).
    #
    domain = os.environ.get("FOURSIGHT_COGNITO_DOMAIN")
    if not domain:
        domain = Gac.get_secret_value("COGNITO_DOMAIN")
    user_pool_id = os.environ.get("FOURSIGHT_COGNITO_USER_POOL_ID")
    if not user_pool_id:
        user_pool_id = Gac.get_secret_value("COGNITO_USER_POOL_ID")
    client_id = os.environ.get("FOURSIGHT_COGNITO_CLIENT_ID")
    if not client_id:
        client_id = Gac.get_secret_value("COGNITO_CLIENT_ID")
    return { # noqa: TypedDict 'CognitoConfig' has missing keys et cetera - using KEY_XYZ constants
        KEY_REGION: AWS_REGION,
        KEY_DOMAIN: domain,
        KEY_USER_POOL_ID: user_pool_id,
        KEY_CLIENT_ID: client_id,
        KEY_SCOPES: ["openid", "email", "profile"],
        KEY_CONNECTIONS: ["Google"],
        KEY_CONFIG_URL: f"{AWS_COGNITO_SERVICE_BASE_URL}/{user_pool_id}/.well-known/openid-configuration",
        KEY_CALLBACK_URL: ""
    }


@memoize
def _get_cognito_oauth_config_client_secret() -> str:
    """
    Returns the Cognito user pool client secret value.
    Note not returned by get_cognito_oauth_config and used only internally to this module.
    :returns: Cognito user pool client secret value.
    """
    client_secret = os.environ.get("FOURSIGHT_COGNITO_CLIENT_SECRET")
    if not client_secret:
        client_secret = Gac.get_secret_value("COGNITO_CLIENT_SECRET")
    return client_secret


def handle_cognito_oauth_callback(request: dict, envs: Envs, site: str,
                                  authtoken_audience: str, authtoken_secret: str) -> dict:
    """
    Called (ultimately) from react_routes for endpoint: GET /cognito/callback
    This is actually called from our primary frontend (React) callback /api/react/cognito/callback
    which is redirected to from Cognito so it can pick up the ouath_pkce_key (sic) which is written
    to browser session storage by the React authentication kickoff code (Amplify.federatedSignIn).
    That value (ouath_pkce_key) is passed to this API as the code_verifier argument, along with the
    code argument which is passed to our primary callback. FYI note known typo in ouath_pkce_key.
    Returns encoded authtoken and expires time suitable for cookie-ing the user on successful login.

    Note that for now at least we use the Auth0 audience (aka client ID) and secret to JWT encode the
    authtoken (for cookie-ing the user on successful login), for straightforward compatibilty with
    existing Auth0 code. I.e. once we've done the initial (login) authentication/authorization we
    act exactly like (as-if) previously implemented Auth0 based authentication.

    :param request: Dictionary with HTTP request for the Cognito authentication callback.
    :param envs: Envs object used to get environment and user info for authorization.
    :param site: Site name (foursight-cgap or foursight-fourfront) used in authtoken.
    :param authtoken_audience: Audience (aka client ID) with which to encode our authtoken JWT.
    :param authtoken_secret: Secret (aka client secret) with which to encode our authtoken JWT.
    :returns: Dictionary with authtoken and expires time which (React) caller will cookie user with.
    """
    # Retrieve (via /oauth2/token) and decode the OAuth (JWT) token, given code/code_verifier arguments.
    token = _retrieve_cognito_oauth_token(request)
    # Create our authtoken (to cookie user) based on the retrieved token.
    authtoken, expires = _create_cognito_authtoken(token, envs, get_request_domain(request), site)
    # Encode our authtoken using the given (Auth0 actually) audience (aka client ID) and secret.
    authtoken_encoded = jwt_encode(authtoken, audience=authtoken_audience, secret=authtoken_secret)
    return {"authtoken": authtoken_encoded, "expires": expires}


def _retrieve_cognito_oauth_token(request: dict) -> dict:
    """
    Calls the /oauth2/token endpoint with the code (and other relevant data from the given
    request, e.g. code_verifier) to retrieve the associated JWT token (id_token) and returns
    its decoded value. See: _call_cognito_oauth_token_endpoint for details on request arguments.

    Cognito configuration dependencies: domain, client ID, client secret.
    Cognito endpoint dependencies: Token i.e. /oauth2/token, JWKS i.e. /.well-known/jwks.json.
    Request arguments: code, code_verifier, state, oauth_state

    This is called from our backend authentication callback (i.e. /api/callback) which itself
    is redirected to from our frontend authentication callback (i.e. /api/react/oauth/callback).
    This is split into two pieces (a frontend and backend part) because we need the frontend
    callback to pick up the PKCE (Proof Key for Code Exchange) value (aka code_verifier) from
    browser local storage (which was written at the start of the authentication process, i.e.
    by Amplify.federatedSignIn), which obviously can only be done by frontend code; and the
    backend is needed to make the /oauth2/token (POST) call to the (Cognito) authentication
    server because it needs to pass to that call the (Cognito) client secret, which obviously
    must be done by backend code so as to not expose this secret outside of that secured context.

    https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce

    :param request: Dictionary with HTTP request for our Cognito authentication callback.
    :returns: Dictionary with decoded JWT token (id_token) from Cognito /oauth2/token endpoint call.
    """
    response = _call_cognito_oauth_token_endpoint(request)
    #
    # Note that we also get back from the /oauth2/token call above (in addition to the "id_token" JWT,
    # which we use) an "access_token" and a "refresh_token"; we do not currently use these; and trying
    # to decode the access_token gives us (unless we disable signature verification) an error because
    # there is no "aud" field there, and trying to decode the refresh_token gives us an invalid payload
    # padding error for some as yet unknown reason, but no matter for now as we don't use these.
    #
    # Note that we also get back from the /oauth2/token call above an "expires_in" (e.g. 3600, for an
    # hour from "now") but the JWT (id_token) also contains a "exp" field (a time_t value) which is
    # effectively the same thing; we will just use the latter (see: _create_cognito_authtoken).
    #
    token = response.get("id_token")
    decoded_token = _decode_cognito_oauth_token_jwt(token)
    return decoded_token


def _call_cognito_oauth_token_endpoint(request: dict) -> dict:
    """
    Calls the Cognito /oauth2/token endpoint to exchange an authorization code for a (JWT) token.
    This authorization "code" is an argument within the given request; this code, also along with
    a "state", was passed to our authentication callback from Cognito. This request also contains
    a "code_verifier" argument, which came from "ouath_pkce_key" (sic) which was stored in browser
    local storage by our client-side authentication initiation code (i.e. Amplify.federatedSignIn),
    together with "oauth_state" which should match the request state argument.

    Cognito configuration dependencies: domain, client ID, client secret.
    Cognito endpoint dependencies: Token i.e. /oauth2/token, JWKS i.e. /.well-known/jwks.json.
    Request arguments: code, code_verifier, state, oauth_state

    :param request: Dictionary with HTTP request for our Cognito authentication callback.
    :returns: Dictionary with HTTP response for Cognito /oauth2/token endpoint (POST) call.
    """
    args = request.get("query_params") or {}
    code = args.get("code")
    #
    # Note the odd/known misspelling of "ouath_pkce_key" browser local storage variable.
    # This, and the above "oauth_state", is set by the client-side code which initiated the
    # authentication process, i.e. e.g. Amplify.federatedSignIn. This is the "code_verifier"
    # which we pass as an argument, along with the given code, to the /oauth2/token endpoint.
    #
    code_verifier = args.get("code_verifier")
    data = _get_cognito_oauth_token_endpoint_data(request, code=code, code_verifier=code_verifier)
    #
    # Note that for the /oauth2/token endpoint call we need EITHER this authorization
    # header, containing the authentication server client secret, OR we need to pass
    # the client secret to the endpoint witin the (POST) data; so BOTH are NOT
    # required; but it is OK to pass both; but we just pass it in the header.
    #
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Authorization": f"Basic {_get_cognito_oauth_token_endpoint_authorization()}"
    }
    url = _get_cognito_oauth_token_endpoint_url()
    cognito_auth_token_response = requests.post(url, headers=headers, data=data)
    if cognito_auth_token_response.status_code != 200:
        raise Exception("Invalid response from /oauth2/token")
    cognito_auth_token_response_json = cognito_auth_token_response.json()
    return cognito_auth_token_response_json


def _get_cognito_oauth_token_endpoint_url() -> str:
    """
    Returns the URL for the POST to the Cognito /oauth2/token endpoint.
    Cognito configuration dependencies: domain.
    :returns: URL for Cognito /oauth2/token endpoint.
    """
    config = _get_cognito_oauth_config_basic()
    domain = config[KEY_DOMAIN]
    return f"https://{domain}/oauth2/token"


def _get_cognito_oauth_token_endpoint_authorization() -> str:
    """
    Returns the value for basic authorization value suitable
    for the header of a POST to the Cognito /oauth2/token endpoint.
    Cognito configuration dependencies: client ID, client secret.
    :returns: Authorization header value for Cognito /oauth2/token endpoint.
    """
    config = _get_cognito_oauth_config_basic()
    client_id = config[KEY_CLIENT_ID]
    client_secret = _get_cognito_oauth_config_client_secret()
    return base64_encode(f"{client_id}:{client_secret}")


def _get_cognito_oauth_token_endpoint_data(request: dict, code: str, code_verifier: str) -> dict:
    """
    Returns the data (dictionary) suitable as the payload for a POST to the /oauth2/token
    endpoint, given an authorization code and associated code_verifier passed to our
    our authentication callback.

    Cognito configuration dependencies: client ID, our authentication callback URL.

    :param code: Value passed from Cognito to our backend authentication callback.
    :param code_verifier: Value passed from Cognito to our backend authentication callback.
    :returns: Data suitable for POST payload for Cognito /oauth2/token endpoint.
    """
    config = get_cognito_oauth_config(request)
    client_id = config[KEY_CLIENT_ID]
    callback = config[KEY_CALLBACK_URL]
    return {
        "grant_type": "authorization_code",
        "client_id": client_id,
        #
        # Note that we do NOT pass the client secret here as we are passing
        # it in the header to the /oauth2/token endpoint (POST) call.
        # See: _get_cognito_oauth_token_endpoint_authorization.
        #
        "code": code,
        "code_verifier": code_verifier,
        "redirect_uri": callback
    }


def _decode_cognito_oauth_token_jwt(jwt: str, verify_signature: bool = True, verify_expiration = True) -> dict:
    """
    Decodes and returns the dictionary for the given JWT (which is assumed to have been returned
    by the Cognito /auth2/token endpoint). To do this we use the signing key within the given JWT
    which we extract using info from the Cognito authentication server JWKA (JSON Web Key Sets) API.

    Note that this is different than how we decode the JWT for Auth0 (from its /oauth/token
    endpoint) which just uses the Auth0 client ID (aka aud) and associated secret.

    Also note that, as alluded to, this JWT is the one returned by the Cognito /oauth2/token endpoint.
    We create/use our own (entirely different) JWT which we use to cookie the user on successful login;
    this one is signed and encoded using the normal client ID (aka aud) and secret mechanism (i.e. not
    a signing key); and we use this same strategy for both Auth0 and Cognito; i.e once a user is
    authenticated via Cognito, the authentication/authorization flow is the same for Cognito and Auth0.

    Cognito configuration dependencies: user pool ID, client ID.
    Cognito endpoint dependencies: JWKS endpoint .../.well-known/jwks.json.

    :param jwt: JWT token value (id_token) retrieved from the /oauth2/token endpoint.
    :param verify_signature: Boolean (default True) indicating JWT signature verification.
    :param verify_expiration: Boolean (default True) indicating JWT expiration verification.
    :returns: Dictionary with decoded value of given JWT.
    """
    #
    # Example decoded (id_tokn) JWT:
    # {
    #     "at_hash": "C9sA39psH6zoRKXuLfQVGw",
    #     "sub": "7dca9b92-746b-4406-9f7d-3e58cd7b247f",
    #     "cognito:groups": [
    #         "us-east-1_h6I5IqQSs_Google"
    #     ],
    #     "email_verified": false,
    #     "iss": "https://cognito-idp.us-east-1.amazonaws.com/us-east-1_h6I5IqQSs",
    #     "cognito:username": "google_117300206013007398924",
    #     "given_name": "David",
    #     "nonce": "Mm9KTFynPpG7J64FA9XxTBYnjmve5S0qjXPLQmGRE-ET-CETERA",
    #     "origin_jti": "04ee7dcf-a2ca-4df5-8790-604456609d69",
    #     "aud": "5d586se3r976435167nk8k8s4h",
    #     "identities": [
    #         {
    #             "userId": "117300206013007398924",
    #             "providerName": "Google",
    #             "providerType": "Google",
    #             "issuer": null,
    #             "primary": "true",
    #             "dateCreated": "1673913190125"
    #         }
    #     ],
    #     "token_use": "id",
    #     "auth_time": 1673913192,
    #     "exp": 1673916792,
    #     "iat": 1673913192,
    #     "family_name": "Michaels",
    #     "jti": "c2061278-c0cd-4bf6-ad77-1ffaeed2f626",
    #     "email": "david_michaels@hms.harvard.edu"
    # }
    #
    config = _get_cognito_oauth_config_basic()
    client_id = config[KEY_CLIENT_ID]
    signing_key = _get_cognito_oauth_signing_key(jwt)
    options = {
        "verify_signature": verify_signature,
        "verify_exp": verify_expiration
    }
    return jwtlib.decode(jwt, signing_key, audience=client_id, algorithms=["RS256"], options=options)


def _get_cognito_oauth_signing_key(jwt: str) -> RSAPublicKey:
    """
    Returns the signing key (object) from the given JWT.
    This is used to verify and decode the JWT (see: _decode_cognito_oauth_token_jwt).
    Actual type of returned object: cryptography.hazmat.backends.openssl.rsa._RSAPublicKey.

    Cognito configuration dependencies: user pool ID.
    Cognito endpoint dependencies: JWKS i.e. /.well-known/jwks.json.

    :param jwt: JWT token value (id_token) retrieved from the /oauth2/token endpoint.
    :returns: Object suitable for use as a signing key to decode a JWT.
    """
    signing_key_client = _get_cognito_oauth_signing_key_client()
    signing_key = signing_key_client.get_signing_key_from_jwt(jwt)
    return signing_key.key


def _get_cognito_oauth_signing_key_client() -> PyJWKClient:
    """
    Returns an object which can be used to extract a signing key from a JWT.
    The returned object will contain a "get_signing_key_from_jwt" method, which takes
    JWT string argument and returns a signing key object which has a "key" property.
    Hits the jwks.json (JSON Web Key Sets) API for the Cognito authentication server.
    Actual type of returned object: jwt.jwks_client.PyJWKClient

    Cognito configuration dependencies: user pool ID.
    Cognito endpoint dependencies: JWKS i.e. /.well-known/jwks.json.

    :returns: Object suitable for extracting a signing key from a JWT.
    """
    config = _get_cognito_oauth_config_basic()
    user_pool_id = config[KEY_USER_POOL_ID]
    cognito_jwks_url = f"{AWS_COGNITO_SERVICE_BASE_URL}/{user_pool_id}/.well-known/jwks.json"
    return PyJWKClient(cognito_jwks_url)


def _create_cognito_authtoken(token: dict, envs: Envs, domain: str, site: str) -> Tuple[dict, int]:
    """
    Creates from the given (decoded) JWT token, retrieved from the /oauth2/token endpoint, an
    authtoken suitable for use as a cookie to indicate the user has been authenticated (logged in).
    This is analagous to foursight_core.react.api.auth.create_authtoken used for Auth0 authentication.

    :param token: Decoded JWT token from the /oauth2/token endpoint.
    :param envs: An Envs object (from react_api_base.ReactApiBase).
    :param domain: Domain name of this instance of the application (i.e. Foursight server itself).
    :param site: Either foursight-cgap or foursignt-fourfront as appropriate.
    :returns: JWT-encoded "authtoken" dictionary suitable for cookie-ing the authenticated user.
    """
    email = token.get("email")
    default_env = envs.get_default_env()
    known_envs = envs.get_known_envs()
    allowed_envs, first_name, last_name = envs.get_user_auth_info(email)
    expires = token.get("exp")
    authtoken = {
        "authentication": "cognito",
        "authenticator": "google",  # TODO: get from identities
        "authenticated_at": token.get("iat"),
        "authenticated_until": expires,
        "authenticated": True,
        "user": email,
        "user_verified": token.get("email_verified"),
        "first_name": token.get("given_name") or first_name,
        "last_name": token.get("family_name") or last_name,
        "allowed_envs": allowed_envs,
        "known_envs": known_envs,
        "default_env": default_env,
        "domain": domain,
        "site": site
    }
    return authtoken, expires


def clear_cognito_cache():
    _get_cognito_oauth_config_basic.cache_clear()
    _get_cognito_oauth_config_client_secret.cache_clear()



from chalice.app import Request
from functools import lru_cache
import inspect
import json
import os
import pkg_resources
import sys
from typing import Callable, Optional, Tuple, Union
from urllib.parse import urlparse

memoize = lru_cache(100)


def sort_dictionary_by_case_insensitive_keys(dictionary: dict) -> dict:
    """
    Returns the given dictionary sorted by (case-insenstivie) key values; yes,
    dictionaries are ordered as of Python 3.7. If the given value is not a
    dictionary it will be coerced to one.
    :param dictionary: Dictionary to sort.
    :return: Given dictionary sorted by key value.
    """
    if not dictionary or not isinstance(dictionary, dict):
        return {}
    return {key: dictionary[key] for key in sorted(dictionary.keys(), key=lambda key: key.lower())}


def get_request_domain(request: dict) -> str:
    return request.get("headers", {}).get("host")


def get_request_args(request: dict) -> dict:
    return request.get("query_params") or {}


def get_request_arg(request: dict, name: str) -> Optional[str]:
    """
    Returns the value of the given URL query parameter name for the given request;
    returns None if not present. Note if present but with no value, then and empty
    string is returned, e.g. for http://localhost:8000/callback?react if "react"
    is queried then returns an empty string, and not None.
    """
    return get_request_args(request).get(name, None)


def get_request_body(request: Request) -> dict:
    return json.loads(request.raw_body.decode())


def get_request_origin(request: dict) -> str:
    headers = request.get("headers", {})
    scheme = headers.get("x-forwarded-proto", "http" if is_running_locally(request) else "https")
    domain = headers.get("host")
    return f"{scheme}://{domain}"


def get_base_url(url: str) -> str:
    """
    Returns the base (root) URL if the given URL, i.e. with just the scheme/protocol (e.g.
    http:// or https://), and the the domain, and port if any. Any credentials (username,
    password) in the URL (e.g. http:) will NOT be included. For example, given this:
    https://username:password@example.com/api/react/cgap-supertest/home?arg=12345678
    this function returns this: https://example.com
    """
    try:
        url = urlparse(url)
        scheme = url.scheme
        hostname = url.hostname
        port = url.port
        port = '' if (port == 80 and scheme == 'http') or (port == 443 and scheme == 'https') else port
        return scheme + "://" + hostname + (":" + str(port) if port else "")
    except Exception:
        return ""


def get_github_url(package: str, file: Optional[str] = None, line: Optional[int] = None) -> Optional[str]:
    """
    Returns the GitHub URL to the given package, with the appropriate/current version.
    Only these packages are supported: foursight-core, foursight-cgap, foursight, dcicutils
    If the optional file argument is given then returns the versionsed GitHub URL to that file.
    This is to provide convenience GitHub links to the foursight (e.g. checks) source code in the UI.
    TODO: Seems like there should be a better more programmatic way to determine this.
    """
    github_url = "https://github.com"
    if package == "foursight_core":
        package_source = "foursight-core"
        package_target = "foursight_core"
        repo_org = "4dn-dcic"
    elif package == "chalicelib_cgap":
        package_source = "foursight-cgap"
        package_target = "chalicelib_cgap"
        repo_org = "dbmi-bgm"
    elif package == "chalicelib_fourfront":
        package_source = "foursight"
        package_target = "chalicelib_fourfront"
        repo_org = "4dn-dcic"
    elif package == "dcicutils":
        package_source = "utils"
        package_target = "dcicutils"
        repo_org = "4dn-dcic"
    else:
        return None
    try:
        version = f"v{pkg_resources.get_distribution(package_source).version}"
    except Exception:
        return None
    repo_url = f"{github_url}/{repo_org}/{package_source}"
    if not file:
        # Note we assume tagged version in GitHub.
        return f"{repo_url}/releases/tag/{version}"
    if os.path.isabs(file):
        path = os.path.normpath(file).split(os.sep)
        for i in range(len(path)):
            if path[i] == package_target:
                if i < len(path) - 1:
                    file = "/".join(path[i + 1:])
                    break
            if file.startswith(os.sep):
                file = file[len(os.sep):]
    line = f"#L{line}" if line > 0 else ""
    return f"{repo_url}/blob/{version}/{package_target}/{file}{line}"


def is_running_locally(request: dict) -> bool:
    """
    Returns true iff the given request indicates that we are running locally (localhost).
    """
    return request.get("context", {}).get("identity", {}).get("sourceIp", "") == "127.0.0.1"


def get_function_info(func: Union[str, Callable]) -> Optional[Tuple[str, str, str, str, int, str]]:
    """
    Returns a tuple containing, in order, these function properties of the given function by
    function name or object: name, file, module, package, line number, GitHub link (only
    if one of these repos: foursight-core, foursight-cgap, foursight, dcicutils)
    Currently used only for informational purposes in the React UI to display
    information about code for checks and actions and their GitHub links.
    """
    def import_function(fully_qualified_function_name: str) -> Optional[Callable]:
        try:
            module_name, unit_name = fully_qualified_function_name.rsplit(".", 1)
            return getattr(__import__(module_name, fromlist=[""]), unit_name)
        except:
            return None
    func_name = None
    func_file = None
    func_module = None
    func_package = None
    func_line = None
    try:
        if not isinstance(func, Callable):
            func = import_function(func)
            if getattr(func, "func"):
                func = func.func
        func_name = func.__name__
        func_module = func.__module__
        func_file = sys.modules[func_module].__file__
        _, func_line = inspect.getsourcelines(func)
        func_package = __import__(func_module).__package__
    except Exception:
        pass
    func_github_url = get_github_url(func_package, func_file, func_line)
    return func_name, func_file, func_module, func_package, func_line, func_github_url


# TODO: Included here until we get utils PR-236 approved/merged/pushed

def keys_and_values_to_dict(keys_and_values: list, key_name: str = "Key", value_name: str = "Value") -> dict:
    """
    Transforms the given list of key/value objects, each containing a "Key" and "Value" property,
    or alternately named via the key_name and/or value_name arguments, into a simple
    dictionary of keys/values, and returns this value. For example, given this:

      [
        { "Key": "env",
          "Value": "prod"
        },
        { "Key": "aws:cloudformation:stack-name",
          "Value": "c4-network-main-stack"
        }
      ]

    This function would return this:

      {
        "env": "prod",
        "aws:cloudformation:stack-name": "c4-network-main-stack"
      }

    :param keys_and_values: List of key/value objects as described above.
    :param key_name: Name of the given key property in the given list of key/value objects; default to "Key".
    :param value_name: Name of the given value property in the given list of key/value objects; default to "Value".
    :returns: Dictionary of keys/values from given list of key/value object as described above.
    :raises ValueError: if item in list does not contain key or value name; or on duplicate key name in list.
    """
    # result = {}
    # for item in keys_and_values:
    #     key = item.get(key_name)
    #     if key:
    #         result[str(key)] = item.get(value_name)
    # return result

    result = {}
    for item in keys_and_values:
        if key_name not in item:
            raise ValueError(f"Key {key_name} is not in {item}.")
        if value_name not in item:
            raise ValueError(f"Key {value_name} is not in {item}.")
        if item[key_name] in result:
            raise ValueError(f"Key {key_name} is duplicated in {keys_and_values}.")
        result[item[key_name]] = item[value_name]
    return result


def dict_to_keys_and_values(dictionary: dict, key_name: str = "Key", value_name: str = "Value") -> list:
    """
    Transforms the keys/values in the given dictionary to a list of key/value objects, each containing
    a "Key" and "Value" property, or alternately named via the key_name and/or value_name arguments,
    and returns this value. For example, given this:

      {
        "env": "prod",
        "aws:cloudformation:stack-name": "c4-network-main-stack"
      }

    This function would return this:

      [
        { "Key": "env",
          "Value": "prod"
        },
        { "Key": "aws:cloudformation:stack-name",
          "Value": "c4-network-main-stack"
        }
      ]

    :param dictionary: Dictionary of keys/values described above.
    :param key_name: Name of the given key property in the result list of key/value objects; default to "Key".
    :param value_name: Name of the given value property in the result list of key/value objects; default to "Value".
    :returns: List of key/value objects from the given dictionary as described above.
    """
    result = []
    for key in dictionary:
        result.append({key_name: key, value_name: dictionary[key]})
    return result



from datetime import datetime, timedelta
import pytz
from dcicutils import ff_utils
from typing import Optional, Tuple
from ...app import app


_PORTAL_ACCESS_KEY_NAME = "access_key_foursight"
_PORTAL_ACCESS_KEY_USER_EMAIL = "foursight.app@gmail.com"
_PORTAL_ACCESS_KEY_EXPIRES_SOON_WARNING_DAYS = 10


def get_portal_access_key_info(env: str,
                               obfuscate: bool = False,
                               test_mode_access_key_simulate_error: bool = False,
                               test_mode_access_key_expiration_warning_days: int = 0) -> dict:
    try:
        now = datetime.now(pytz.utc)
        connection = app.core.init_connection(env)
        connection_keys = connection.ff_keys
        key = connection_keys.get("key")
        if key and obfuscate:
            key = key[0] + "*******"
        secret = connection_keys.get("secret")
        if secret:
            secret = "********" if obfuscate else secret[1] + "*******"
        server = connection_keys.get("server")
        access_key_info = {"key": key, "secret": secret, "server": server}
        access_key_create_date, access_key_expires_date, access_key_expires_exception = \
            _get_portal_access_key_expires_date(connection_keys)
        if access_key_expires_date:
            access_key_info["created_at"] = access_key_create_date.strftime("%Y-%m-%d %H:%M:%S")
            if access_key_expires_date == datetime.max:
                access_key_info["expires_at"] = None
                access_key_info["expired"] = False
                access_key_info["invalid"] = False
                access_key_info["expires_soon"] = False
            else:
                access_key_info["expires_at"] = access_key_expires_date.strftime("%Y-%m-%d %H:%M:%S")
                access_key_info["expired"] = now >= access_key_expires_date
                access_key_info["invalid"] = access_key_info["expired"]
                if test_mode_access_key_expiration_warning_days > 0:
                    expires_soon_days = test_mode_access_key_expiration_warning_days
                else:
                    expires_soon_days = _PORTAL_ACCESS_KEY_EXPIRES_SOON_WARNING_DAYS
                access_key_info["expires_soon"] = _is_datetime_within_future_ndays(access_key_expires_date,
                                                                                   ndays=expires_soon_days,
                                                                                   now=now)
            if test_mode_access_key_simulate_error:
                access_key_info["exception"] = "test_mode_access_key_simulate_error"
                access_key_info["invalid"] = True
        else:
            # If we don't get an expires date at all, then assume there is a problem.
            access_key_info["exception"] = str(access_key_expires_exception)
            e = str(access_key_info["exception"]).lower()
            if "credenti" in e or "expir" in e:
                access_key_info["probably_expired"] = True
            access_key_info["invalid"] = True
        access_key_info["timestamp"] = now.strftime("%Y-%m-%d %H:%M:%S")
        return access_key_info
    except Exception:
        return {}


def _get_portal_access_key_expires_date(keys: dict) -> Tuple[datetime, Optional[datetime], Optional[Exception]]:
    try:
        query = f"/search/?type=AccessKey&description={_PORTAL_ACCESS_KEY_NAME}&sort=-date_created"
        access_key = ff_utils.search_metadata(query, key=keys)[0]
        access_key_create_date = datetime.fromisoformat(access_key["date_created"]).astimezone(pytz.utc)
        access_key_expires_date = access_key.get("expiration_date")
        if access_key_expires_date:
            access_key_expires_date = datetime.fromisoformat(access_key_expires_date).astimezone(pytz.utc)
        else:
            # There may or may not be an expiration date (e.g. for fourfront);
            # if not then make it the max date which is 9999-12-31.
            access_key_expires_date = datetime.max
        return (access_key_create_date, access_key_expires_date, None)
    except Exception as e:
        return (None, None, e)


def _is_datetime_within_future_ndays(d: datetime, ndays: int, now: Optional[datetime] = None) -> bool:
    """
    Returns True iff the given datatime is within ndays IN THE FUTURE of the CURRENT
    datetime. I.e. if the current datetime plus ndays is greater-than-or-equal to
    the given datetime the returns True; otherwise returns False.
    """
    if not now:
        now = datetime.now(pytz.utc)
    return (d - now) <= timedelta(days=ndays) if d >= now else False



import datetime
from typing import Any, Optional
from http.cookies import SimpleCookie
from dcicutils.misc_utils import str_to_bool
from .datetime_utils import (
    convert_time_t_to_datetime,
    convert_utc_datetime_to_cookie_expires_format,
    EPOCH
)
from .misc_utils import is_running_locally


def read_cookie(request: dict, cookie_name: str) -> Optional[str]:
    """
    Returns the value of the cookie of the given name from within the given request,
    or None if it does not exist.
    """
    if not cookie_name or not request:
        return ""
    simple_cookies = _read_cookies(request)
    cookie_morsel = simple_cookies.get(cookie_name) if simple_cookies else None
    return cookie_morsel.value if cookie_morsel else None


def read_cookie_bool(request: dict, cookie_name: str) -> bool:
    return str_to_bool(read_cookie(request, cookie_name))


def read_cookie_int(request: dict, cookie_name: str, fallback: int = 0) -> int:
    value = read_cookie(request, cookie_name)
    if not value:
        return fallback
    if value.startswith("-"):
        value = value[1:]
        multiplier = -1
    else:
        multiplier = 1
    if not value.isdigit():
        return fallback
    return int(value) * multiplier


def _read_cookies(request: dict) -> Optional[SimpleCookie]:
    """
    Returns the cookies from the given request as a SimpleCookie object.
    """
    if not request:
        return None
    cookies = request.get("headers", {}).get("cookie")
    if not cookies:
        return None
    simple_cookies = SimpleCookie()
    simple_cookies.load(cookies)
    return simple_cookies


def create_set_cookie_string(request: dict, name: str, value: Optional[str],
                             domain: str, path: str = "/", expires: Optional[Any] = None,
                             http_only: bool = False) -> str:
    """
    Returns a string suitable for an HTTP response to set a cookie for this given cookie info.
    If the given expires arg is "now" then the expiration time for the cookie will be set
    to the epoch (i.e. 1970-01-01 00:00:00) indicating it has expired; used to delete cookie.
    """
    if not name or not request:
        return ""
    cookie = name + "=" + (value if value else "") + ";"
    if domain and not is_running_locally(request):
        # N.B. When running on localhost cookies cannot be set unless we leave off the domain entirely.
        # https://stackoverflow.com/questions/1134290/cookies-on-localhost-with-explicit-domain
        cookie += f" Domain={domain};"
    if not path:
        path = "/"
    cookie += f" Path={path};"
    if expires:
        if isinstance(expires, datetime.datetime):
            expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        elif isinstance(expires, int):
            expires = convert_time_t_to_datetime(expires)
            expires = expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
        elif isinstance(expires, str):
            if expires.lower() == "now":
                expires = f"Expires={convert_utc_datetime_to_cookie_expires_format(EPOCH)}"
        else:
            expires = None
        if expires:
            cookie += f" Expires={expires};"
    if http_only:
        cookie += " HttpOnly;"
    return cookie


def create_delete_cookie_string(request: dict, name: str, domain: str, path: str = "/") -> str:
    return create_set_cookie_string(request, name=name, value=None, domain=domain, path=path, expires="now")



import yaml


def load_yaml(yaml_string: str) -> dict:
    """
    Loads the given YAML string and returns it as a dictionary.
    If an error occurs on load then and empty dictionary is returned.
    """
    try:
        return yaml.load(yaml_string, Loader=yaml.Loader)
    except Exception:
        return {"error": "Cannot parse YAML."}


class _SpecialTag(yaml.YAMLObject):
    yaml_tag = None
    def __init__(self, val):
        self.val = val
    @classmethod
    def from_yaml(cls, loader, node):
        return str(node.value)


class _Ref(_SpecialTag):
    yaml_tag = '!Ref'


class _Sub(_SpecialTag):
    yaml_tag = '!Sub'


class _Join(_SpecialTag):
    yaml_tag = '!Join'


class _GetAtt(_SpecialTag):
    yaml_tag = '!GetAtt'


class _ImportValue(_SpecialTag):
    yaml_tag = '!ImportValue'
